import os
import click
import shutil
def check_if_path_exists(path):
    if os.path.exists(path):
        if not click.confirm('Cnvrg will override {project} Do you want to continue?'.format(project=path)):
            return
        else:
            shutil.rmtree(path, ignore_errors=True)


from cnvrg.modules.base_module import CnvrgBase
from cnvrg.helpers.apis_helper import post as apis_post, get as apis_get, download_file
from cnvrg.modules.cnvrg_files import CnvrgFiles
from cnvrg.modules.errors import NotImplementedError
from cnvrg.modules import UnknownStsError
from cnvrg.helpers.crypto_helpers import decrypt
import os
from typing import Dict, List, Callable
import click
from cnvrg.helpers.parallel_helper import safe_parallel
from cnvrg.helpers import log_error


class Storage(CnvrgBase):
    def __init__(self, element: CnvrgFiles, working_dir: str, sts_path: str):
        self.element = element
        self.working_dir = working_dir
        self.init_sts(sts_path)
        self.conflicts = []

    def init_sts(self, sts_path):
        sts_content = download_file(sts_path)
        if not sts_content:
            raise UnknownStsError("Cant find sts")
        self.key, self.iv = sts_content.split("\n")

    def decrypt(self, text):
        return decrypt(self.key, self.iv, text)

    def decrypt_dict(self, props: Dict, keys: List = None):
        return {k: decrypt(self.key, self.iv, v) if k in keys else v for k,v in props.items()}

    def status(self):
        pass


    def download_file(self, local_path: str, storage_path: str):
        raise NotImplementedError

    def clone_commit(self, commit, **kwargs):
        self.download_files(self.element.fetch_commit(commit).get("blobs"))

    def download_single_file(self, file):
        local_p, storage_p = self.decrypt(file["local_path"]), self.decrypt(file["storage_path"])
        if local_p in self.conflicts:
            local_p = local_p + ".conflict"
        local_p = os.path.join(self.working_dir, local_p)
        self.get_dir(local_p)
        ##if file, download it
        if not local_p.endswith("/"): self.download_file(local_p, storage_p)

    def upload_single_file(self, file):
        local_p, storage_p = self.decrypt(file["local_path"]), self.decrypt(file["storage_path"])
        local_p = os.path.join(self.working_dir, local_p)
        self.upload_file(local_p, storage_p)


    def download_dirs(self, dirs: List):
        try:
            safe_parallel(self.download_single_file, dirs)
        except Exception as e:
            log_error(e)
            raise e

    def download_files(self, files: List, conflicts: List=None):
        try:
            self.conflicts = conflicts or []
            safe_parallel(self.download_single_file, files, progressbar={"desc": "Downloading files"})
            self.conflicts = []
        except Exception as e:
            log_error(e)
            click.secho("Cant download all files")
            raise e

    def delete_files(self, files: List, conflicts: List=None):
        conflicts = conflicts or []
        for file in files:
            if file in conflicts:
                src_name = os.path.join(self.working_dir, file)
                dest_name = src_name + ".deleted"
                os.rename(src_name, dest_name)
            else:
                os.remove(os.path.join(self.working_dir, file))

    def get_dir(self, filename):
        dirname = os.path.dirname(filename)
        if os.path.exists(dirname):
            return
        os.makedirs(dirname, exist_ok=True)

    def upload_files(self, files: List):
        try:
            safe_parallel(self.upload_single_file, files, progressbar={"desc": "Uploading files"})
        except Exception as e:
            click.secho("Cant upload all files")
            raise e


    def upload_file(self, local_path: str, storage_path: str):
        raise NotImplementedError






import sys
import io
import time
import threading

def before_run():
    buffer = io.StringIO()
    old_buffer = sys.stdout
    sys.stdout = buffer
    return buffer, old_buffer


def after_run(buffer, old_buffer):
    buffer.close()
    sys.stdout = old_buffer


def stdout_thread(stdout_buffer, faketdout_buffer, callback):
    def p(seek):
        v = stdout_buffer.getvalue()
        v = v[seek:]
        v = v.replace('[0m', "")
        if v:
            faketdout_buffer.write(v)
            lines = list(filter(lambda x: x, v.strip().split("\n")))
            callback(lines)
        return seek + len(v)

    seek = 0
    while not stdout_buffer.closed:
        seek = p(seek)
        time.sleep(0.5)



def run_callable(callable, callback):
    buffer, old_buffer = before_run()
    t = threading.Thread(target=stdout_thread, args=(buffer, old_buffer, callback))
    t.start()
    exit_status = 0
    try:
        callable()
    except Exception as e:
        callback(e)
        exit_status = 1
    time.sleep(0.5)
    after_run(buffer,old_buffer)
    t.join()
    return exit_status




from Crypto.Cipher import AES
from base64 import b64decode
unpad = lambda s: s[:-ord(s[len(s) - 1:])]


def decrypt(key: str, iv: str, secret: str):
    cipher = AES.new(key.encode("utf8"), AES.MODE_CBC, b64decode(iv.encode("utf8")))
    secret = b64decode(secret.encode("utf-8"))
    return unpad(cipher.decrypt(secret).decode('utf8'))


from cnvrg.modules.base_module import CnvrgBase
from cnvrg.modules.project import Project
import cnvrg.helpers.apis_helper as apis_helper
import cnvrg.helpers.string_helper as string_helper
import os
import cnvrg.helpers.env_helper as env_helper
import cnvrg.helpers.config_helper as config_helper
import cnvrg.helpers.time_helper as time_helper
from cnvrg.modules.errors import UserError
from cnvrg.helpers.url_builder_helper import url_join



LOGS_TYPE_OUTPUT = "output"
LOGS_TYPE_ERROR = "error"
LOGS_TYPE_INFO = "info"
LOGS_TYPE_WARNING = "warning"
MAX_LOGS_PER_SEND = 400

class CnvrgJob(CnvrgBase):
    def __init__(self, slug=None, job_type=None, project=None):
        self.job_slug = slug or env_helper.CURRENT_JOB_ID
        self.job_type = job_type or env_helper.CURRENT_JOB_TYPE
        if not project:
            if not config_helper.config_type() == config_helper.CONFIG_TYPE_PROJECT:
                raise UserError(
                    "Cant create an experiment without a project, please pass a project or cd into cnvrg project dir")
            project = Project()
        self.project = project


    def _base_job_url(self):
        return url_join(
            #### hackish :D
            self.project.get_base_url(),"jobs", string_helper.to_snake_case(self.job_type), self.job_slug
        )


    def send_util(self, utilization):
        apis_helper.post(url_join(self._base_job_url(), "utilization"), data=[utilization])


    def log(self, logs, log_type=LOGS_TYPE_OUTPUT):
        if isinstance(logs, str): logs = [logs]
        for i in range(0, len(logs), env_helper.MAX_LOGS_PER_SEND):
            apis_helper.post(url_join(self._base_job_url(), "log"), data={"logs": logs[i:i + env_helper.MAX_LOGS_PER_SEND], "log_level": log_type, "timestamp": time_helper.now_as_string()})
